# Copyright (C) 2024,2025 Kian-Meng Ang
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Watermark subcommand."""

import argparse
import logging
import math

from PIL import Image, ImageColor, ImageDraw, ImageFont, ImageSequence

from fotolab import save_image
from fotolab.subcommands.info import camera_metadata

log = logging.getLogger(__name__)

FONT_SIZE_ASPECT_RATIO = 12 / 600
FONT_PADDING_ASPECT_RATIO = 15 / 600
FONT_OUTLINE_WIDTH_ASPECT_RATIO = 2 / 600
POSITIONS = ["top-left", "top-right", "bottom-left", "bottom-right"]


def build_subparser(subparsers) -> None:
    """Build the subparser."""
    watermark_parser = subparsers.add_parser(
        "watermark", help="watermark an image"
    )

    watermark_parser.set_defaults(func=run)

    watermark_parser.add_argument(
        dest="image_filenames",
        help="set the image filenames",
        nargs="+",
        type=str,
        default=None,
        metavar="IMAGE_FILENAMES",
    )

    watermark_parser.add_argument(
        "-t",
        "--text",
        dest="text",
        help="set the watermark text (default: '%(default)s')",
        type=str,
        default="kianmeng.org",
        metavar="WATERMARK_TEXT",
    )

    watermark_parser.add_argument(
        "-p",
        "--position",
        dest="position",
        choices=POSITIONS,
        help="set position of the watermark text (default: '%(default)s')",
        default="bottom-left",
    )

    watermark_parser.add_argument(
        "-pd",
        "--padding",
        dest="padding",
        type=int,
        default=15,
        help=(
            "set the padding of the watermark text relative to the image "
            "(default: '%(default)s')"
        ),
        metavar="PADDING",
    )

    watermark_parser.add_argument(
        "-fs",
        "--font-size",
        dest="font_size",
        type=int,
        default=12,
        help="set the font size of watermark text (default: '%(default)s')",
        metavar="FONT_SIZE",
    )

    watermark_parser.add_argument(
        "-fc",
        "--font-color",
        dest="font_color",
        type=str,
        default="white",
        help="set the font color of watermark text (default: '%(default)s')",
        metavar="FONT_COLOR",
    )

    watermark_parser.add_argument(
        "-ow",
        "--outline-width",
        dest="outline_width",
        type=int,
        default=2,
        help=(
            "set the outline width of the watermark text "
            "(default: '%(default)s')"
        ),
        metavar="OUTLINE_WIDTH",
    )

    watermark_parser.add_argument(
        "-oc",
        "--outline-color",
        dest="outline_color",
        type=str,
        default="black",
        help=(
            "set the outline color of the watermark text "
            "(default: '%(default)s')"
        ),
        metavar="OUTLINE_COLOR",
    )

    watermark_parser.add_argument(
        "--camera",
        default=False,
        action="store_true",
        dest="camera",
        help="use camera metadata as watermark",
    )

    watermark_parser.add_argument(
        "-l",
        "--lowercase",
        default=True,
        action=argparse.BooleanOptionalAction,
        dest="lowercase",
        help="lowercase the watermark text",
    )


def run(args: argparse.Namespace) -> None:
    """Run watermark subcommand.

    Args:
        args (argparse.Namespace): Config from command line arguments

    Returns:
        None
    """
    log.debug(args)

    for image_filename in args.image_filenames:
        image = Image.open(image_filename)
        if image.format == "GIF":
            watermark_gif_image(image, args)
        else:
            watermarked_image = watermark_non_gif_image(image, args)
            save_image(args, watermarked_image, image_filename, "watermark")


def watermark_gif_image(original_image: Image.Image, args: argparse.Namespace):
    """Watermark the image."""
    frames = []
    for frame in ImageSequence.Iterator(original_image):
        watermarked_frame = watermark_image(args, frame.convert("RGBA"))
        frames.append(watermarked_frame)

    frames[0].save(
        "foo.gif",
        format="GIF",
        append_images=frames[1:],
        save_all=True,
        duration=original_image.info.get("duration", 100),
        loop=original_image.info.get("loop", 0),
        disposal=original_image.info.get("disposal", 2),
    )


def watermark_non_gif_image(
    original_image: Image.Image, args: argparse.Namespace
):
    """Watermark the image."""
    return watermark_image(args, original_image)


def watermark_image(
    args: argparse.Namespace, original_image: Image.Image
) -> Image.Image:
    """Watermark an image."""
    watermarked_image = original_image.copy()
    draw = ImageDraw.Draw(watermarked_image)

    font = ImageFont.load_default(calc_font_size(original_image, args))
    log.debug("default font: %s", " ".join(font.getname()))

    text = prepare_text(args, original_image)
    (left, top, right, bottom) = draw.textbbox(xy=(0, 0), text=text, font=font)
    text_width = right - left
    text_height = bottom - top
    (position_x, position_y) = calc_position(
        watermarked_image,
        text_width,
        text_height,
        args.position,
        calc_padding(original_image, args),
    )

    draw.text(
        (position_x, position_y),
        text,
        font=font,
        fill=(*ImageColor.getrgb(args.font_color), 128),
        stroke_width=calc_font_outline_width(original_image, args),
        stroke_fill=(*ImageColor.getrgb(args.outline_color), 128),
    )
    return watermarked_image


def prepare_text(args, image) -> str:
    """Prepare the watermark text."""
    text = args.text
    if args.camera and camera_metadata(image):
        text = camera_metadata(image)

    if args.lowercase:
        text = text.lower()

    return text


def calc_font_size(image, args) -> int:
    """Calculate the font size based on the width of the image."""
    width, _height = image.size
    new_font_size = args.font_size
    if width > 600:
        new_font_size = math.floor(FONT_SIZE_ASPECT_RATIO * width)

    log.debug("new font size: %d", new_font_size)
    return new_font_size


def calc_font_outline_width(image, args) -> int:
    """Calculate the font padding based on the width of the image."""
    width, _height = image.size
    new_font_outline_width = args.outline_width
    if width > 600:
        new_font_outline_width = math.floor(
            FONT_OUTLINE_WIDTH_ASPECT_RATIO * width
        )

    log.debug("new font outline width: %d", new_font_outline_width)
    return new_font_outline_width


def calc_padding(image, args) -> int:
    """Calculate the font padding based on the width of the image."""
    width, _height = image.size
    new_padding = args.padding
    if width > 600:
        new_padding = math.floor(FONT_PADDING_ASPECT_RATIO * width)

    log.debug("new padding: %d", new_padding)
    return new_padding


def calc_position(image, text_width, text_height, position, padding) -> tuple:
    """Calculate the boundary coordinates of the watermark text."""
    positions = {
        "top-left": (padding, padding),
        "top-right": (image.width - text_width - padding, padding),
        "bottom-left": (padding, image.height - text_height - padding),
        "bottom-right": (
            image.width - text_width - padding,
            image.height - text_height - padding,
        ),
    }

    return positions.get(position, (0, 0))



# Copyright (C) 2024,2025 Kian-Meng Ang
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Halftone subcommand."""

import argparse
import logging
import math
from pathlib import Path

from PIL import Image, ImageDraw

from fotolab import _open_image, save_image

log = logging.getLogger(__name__)


def build_subparser(subparsers) -> None:
    """Build the subparser."""
    halftone_parser = subparsers.add_parser(
        "halftone", help="halftone an image"
    )

    halftone_parser.set_defaults(func=run)

    halftone_parser.add_argument(
        dest="image_filenames",
        help="set the image filename",
        nargs="+",
        type=str,
        default=None,
        metavar="IMAGE_FILENAMES",
    )

    halftone_parser.add_argument(
        "-ba",
        "--before-after",
        default=False,
        action="store_true",
        dest="before_after",
        help="generate a GIF showing before and after changes",
    )


def run(args: argparse.Namespace) -> None:
    """Run halftone subcommand.

    Args:
        config (argparse.Namespace): Config from command line arguments

    Returns:
        None
    """
    log.debug(args)

    for image_filename in args.image_filenames:
        original_image = Image.open(image_filename)
        halftone_image = create_halftone_image(original_image)

        if args.before_after:
            save_gif_image(
                args, image_filename, original_image, halftone_image
            )
        else:
            save_image(args, halftone_image, image_filename, "halftone")


def create_halftone_image(
    original_image: Image.Image, cell_count: int = 50
) -> Image.Image:
    """Create a halftone version of the input image.

    Modified from the circular halftone effect processing.py example from
    https://tabreturn.github.io/code/processing/python/2019/02/09/processing.py_in_ten_lessons-6.3-_halftones.html

    Args:
        original_image: The source image to convert
        cell_count: Number of cells across the width (default: 50)

    Returns:
        Image.Image: The halftone converted image
    """
    grayscale_image = original_image.convert("L")
    width, height = original_image.size

    halftone_image = Image.new("L", (width, height), "black")
    draw = ImageDraw.Draw(halftone_image)

    cellsize = width / cell_count
    rowtotal = math.ceil(height / cellsize)

    for row in range(rowtotal):
        for col in range(cell_count):
            # Calculate center point of current cell
            x = int(col * cellsize + cellsize / 2)
            y = int(row * cellsize + cellsize / 2)

            # Get brightness and calculate dot size
            brightness = grayscale_image.getpixel((x, y))
            dot_size = 10 * brightness / 200

            # Draw the dot
            draw.ellipse(
                [
                    x - dot_size / 2,
                    y - dot_size / 2,
                    x + dot_size / 2,
                    y + dot_size / 2,
                ],
                fill=255,
            )

    return halftone_image


def save_gif_image(args, image_filename, original_image, sharpen_image):
    """Save the original and sharpen image."""
    image_file = Path(image_filename)
    new_filename = Path(
        args.output_dir,
        image_file.with_name(f"sharpen_gif_{image_file.stem}.gif"),
    )
    new_filename.parent.mkdir(parents=True, exist_ok=True)

    log.info("sharpen gif image: %s", new_filename)
    original_image.save(
        new_filename,
        format="gif",
        append_images=[sharpen_image],
        save_all=True,
        duration=1000,
        loop=0,
        optimize=True,
    )

    if args.open:
        _open_image(new_filename)



# Copyright (C) 2024,2025 Kian-Meng Ang
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Montage subcommand."""

import argparse
import logging

from PIL import Image

from fotolab import save_image

log = logging.getLogger(__name__)


def build_subparser(subparsers) -> None:
    """Build the subparser."""
    montage_parser = subparsers.add_parser(
        "montage", help="montage a list of image"
    )

    montage_parser.set_defaults(func=run)

    montage_parser.add_argument(
        dest="image_filenames",
        help="set the image filenames",
        nargs="+",
        type=argparse.FileType("r"),
        default=None,
        metavar="IMAGE_FILENAMES",
    )


def run(args: argparse.Namespace) -> None:
    """Run montage subcommand.

    Args:
        config (argparse.Namespace): Config from command line arguments

    Returns:
        None
    """
    log.debug(args)
    images = []
    for image_filename in args.image_filenames:
        images.append(Image.open(image_filename.name))

    total_width = sum(img.width for img in images)
    total_height = max(img.height for img in images)

    montaged_image = Image.new("RGB", (total_width, total_height))

    x_offset = 0
    for image in images:
        montaged_image.paste(image, (x_offset, 0))
        x_offset += image.width

    output_image_filename = args.image_filenames[0].name
    save_image(args, montaged_image, output_image_filename, "montage")



# Copyright (C) 2024,2025 Kian-Meng Ang

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Generals Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A console program to manipulate photos.

  website: https://github.com/kianmeng/fotolab
  changelog: https://github.com/kianmeng/fotolab/blob/master/CHANGELOG.md
  issues: https://github.com/kianmeng/fotolab/issues
"""

import argparse
import logging
import sys
from typing import Optional, Sequence

import fotolab.subcommands
from fotolab import __version__

log = logging.getLogger(__name__)


def setup_logging(args: argparse.Namespace) -> None:
    """Sets up logging configuration based on command-line arguments.

    Args:
        args (argparse.Namespace): Namespace containing parsed arguments.
    """
    if args.quiet:
        logging.disable(logging.NOTSET)
        return

    if args.verbose == 0:
        logging.getLogger("PIL").setLevel(logging.ERROR)

    level = logging.DEBUG if args.debug else logging.INFO
    format_string = (
        "[%(asctime)s] %(levelname)s: %(name)s: %(message)s"
        if args.debug
        else "%(message)s"
    )

    logging.basicConfig(
        level=level,
        format=format_string,
        stream=sys.stdout,
        datefmt="%Y-%m-%d %H:%M:%S",
    )


def build_parser() -> argparse.ArgumentParser:
    """Build the CLI parser."""
    parser = argparse.ArgumentParser(
        prog="fotolab",
        description=__doc__,
        formatter_class=lambda prog: argparse.RawTextHelpFormatter(
            prog, max_help_position=20
        ),
    )

    parser.add_argument(
        "-o",
        "--overwrite",
        default=False,
        action="store_true",
        dest="overwrite",
        help="overwrite existing image",
    )

    parser.add_argument(
        "-op",
        "--open",
        default=False,
        action="store_true",
        dest="open",
        help="open the image using default program (default: '%(default)s')",
    )

    parser.add_argument(
        "-od",
        "--output-dir",
        dest="output_dir",
        default="output",
        help="set default output folder (default: '%(default)s')",
    )

    parser.add_argument(
        "-q",
        "--quiet",
        default=False,
        action="store_true",
        dest="quiet",
        help="suppress all logging",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        default=0,
        action="count",
        dest="verbose",
        help="show verbosity of debugging log, use -vv, -vvv for more details",
    )

    parser.add_argument(
        "-d",
        "--debug",
        default=False,
        action="store_true",
        dest="debug",
        help="show debugging log and stacktrace",
    )

    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
    )

    subparsers = parser.add_subparsers(help="sub-command help")
    fotolab.subcommands.build_subparser(subparsers)

    return parser


def main(args: Optional[Sequence[str]] = None) -> None:
    """Run the main program flow."""
    args = args or sys.argv[1:]
    log.debug(args)

    try:
        parser = build_parser()
        if len(args) == 0:
            parser.print_help(sys.stderr)
        else:
            parsed_args = parser.parse_args(args)
            setup_logging(parsed_args)

            if hasattr(parsed_args, "func"):
                log.debug(parsed_args)
                parsed_args.func(parsed_args)
            else:
                parser.print_help(sys.stderr)

    except Exception as error:
        log.error(
            "error: %s",
            getattr(error, "message", str(error)),
            exc_info=("-d" in args or "--debug" in args),
        )

        raise SystemExit(1) from None



# Copyright (C) 2024,2025 Kian-Meng Ang
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Rotate subcommand."""

import argparse
import logging

from PIL import Image

from fotolab import save_image

log = logging.getLogger(__name__)


def build_subparser(subparsers) -> None:
    """Build the subparser."""
    rotate_parser = subparsers.add_parser("rotate", help="rotate an image")

    rotate_parser.set_defaults(func=run)

    rotate_parser.add_argument(
        dest="image_filenames",
        help="set the image filenames",
        nargs="+",
        type=str,
        default=None,
        metavar="IMAGE_FILENAMES",
    )

    rotate_parser.add_argument(
        "-r",
        "--rotation",
        type=int,
        default=0,
        help="Rotation angle in degrees (default: '%(default)s')",
    )

    rotate_parser.add_argument(
        "-cw",
        "--clockwise",
        action="store_true",
        help="Rotate clockwise (default: '%(default)s)",
    )


def run(args: argparse.Namespace) -> None:
    """Run rotate subcommand.

    Args:
        config (argparse.Namespace): Config from command line arguments

    Returns:
        None
    """
    log.debug(args)

    rotation = args.rotation
    if args.clockwise:
        rotation = -rotation

    for image_filename in args.image_filenames:
        original_image = Image.open(image_filename)
        rotated_image = original_image.rotate(
            rotation,
            expand=True,
        )
        save_image(args, rotated_image, image_filename, "rotate")
